@@ -22,10 +22,12 @@
 # FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 # License for more details.
 
+import re
 import os
 import sys
 import time
 import threading
+import subprocess as sp
 from operator import attrgetter
 
 import psycopg2
@@ -33,7 +35,7 @@ import psycopg2.errorcodes
 from psycopg2 import extensions as ext
 
 from testutils import (
-    unittest, decorate_all_tests, skip_if_no_superuser,
+    script_to_py3, unittest, decorate_all_tests, skip_if_no_superuser,
     skip_before_postgres, skip_after_postgres, skip_before_libpq,
     ConnectingTestCase, skip_if_tpc_disabled, skip_if_windows, slow)
 
@@ -72,16 +74,8 @@ class ConnectionTests(ConnectingTestCase):
         # ticket #148
         conn = self.conn
         cur = conn.cursor()
-        try:
-            cur.execute("select pg_terminate_backend(pg_backend_pid())")
-        except psycopg2.OperationalError, e:
-            if e.pgcode != psycopg2.errorcodes.ADMIN_SHUTDOWN:
-                raise
-        except psycopg2.DatabaseError, e:
-            # curiously when disconnected in green mode we get a DatabaseError
-            # without pgcode.
-            if e.pgcode is not None:
-                raise
+        self.assertRaises(psycopg2.OperationalError,
+            cur.execute, "select pg_terminate_backend(pg_backend_pid())")
 
         self.assertEqual(conn.closed, 2)
         conn.close()
@@ -1487,6 +1481,101 @@ class AutocommitTests(ConnectingTestCase):
         self.assertEqual(cur.fetchone()[0], 'on')
 
 
+class PasswordLeakTestCase(ConnectingTestCase):
+    def setUp(self):
+        super(PasswordLeakTestCase, self).setUp()
+        PasswordLeakTestCase.dsn = None
+
+    class GrassingConnection(ext.connection):
+        """A connection snitching the dsn away.
+
+        This connection passes the dsn to the test case class even if init
+        fails (e.g. connection error). Test that we mangle the dsn ok anyway.
+        """
+
+        def __init__(self, *args, **kwargs):
+            try:
+                super(PasswordLeakTestCase.GrassingConnection, self).__init__(
+                    *args, **kwargs)
+            finally:
+                # The connection is not initialized entirely, however the C
+                # code should have set the dsn, and it should have scrubbed
+                # the password away
+                PasswordLeakTestCase.dsn = self.dsn
+
+    def test_leak(self):
+        self.assertRaises(psycopg2.DatabaseError,
+            self.GrassingConnection, "dbname=nosuch password=whateva")
+        self.assertDsnEqual(self.dsn, "dbname=nosuch password=xxx")
+
+    @skip_before_libpq(9, 2)
+    def test_url_leak(self):
+        self.assertRaises(psycopg2.DatabaseError,
+            self.GrassingConnection,
+            "postgres://someone:whateva@localhost/nosuch")
+
+        self.assertDsnEqual(self.dsn,
+            "user=someone password=xxx host=localhost dbname=nosuch")
+
+
+class SignalTestCase(ConnectingTestCase):
+    @slow
+    @skip_before_postgres(8, 2)
+    def test_bug_551_returning(self):
+        # Raise an exception trying to decode 'id'
+        self._test_bug_551(query="""
+            INSERT INTO test551 (num) VALUES (%s) RETURNING id
+            """)
+
+    @slow
+    def test_bug_551_no_returning(self):
+        # Raise an exception trying to decode 'INSERT 0 1'
+        self._test_bug_551(query="""
+            INSERT INTO test551 (num) VALUES (%s)
+            """)
+
+    def _test_bug_551(self, query):
+        script = ("""\
+import os
+import sys
+import time
+import signal
+import threading
+
+import psycopg2
+
+def handle_sigabort(sig, frame):
+    sys.exit(1)
+
+def killer():
+    time.sleep(0.5)
+    os.kill(os.getpid(), signal.SIGABRT)
+
+signal.signal(signal.SIGABRT, handle_sigabort)
+
+conn = psycopg2.connect(%(dsn)r)
+
+cur = conn.cursor()
+
+cur.execute("create table test551 (id serial, num varchar(50))")
+
+t = threading.Thread(target=killer)
+t.daemon = True
+t.start()
+
+while True:
+    cur.execute(%(query)r, ("Hello, world!",))
+""" % {'dsn': dsn, 'query': query})
+
+        proc = sp.Popen([sys.executable, '-c', script_to_py3(script)],
+            stdout=sp.PIPE, stderr=sp.PIPE)
+        (out, err) = proc.communicate()
+        self.assertNotEqual(proc.returncode, 0)
+        # Strip [NNN refs] from output
+        err = re.sub(br'\[[^\]]+\]', b'', err).strip()
+        self.assert_(not err, err)
+
+
 def test_suite():
     return unittest.TestLoader().loadTestsFromName(__name__)
 
